import typing as t

import pydantic


class Text: pass


class Image: pass


class Ref: pass


class File: pass


class LocalFile: pass


T = t.TypeVar('T')


class Label(pydantic.BaseModel):
    name: str
    value: t.Optional[str] = None


# There are order of parent classes required
class Object(pydantic.BaseModel, t.Generic[T]):

    id: str
    project_id: t.Optional[str] = None
    labels: t.Optional[list[Label]] = None
    tp: str
    data: T


class Context(pydantic.BaseModel):
    id: str
    project_id: t.Optional[str] = None
    labels: t.Optional[list[Label]] = None
    tp: str
    # data: T

    def update(self, push={}, set={}):
        return {
            "$context_update": {
                "$id": self.id,
                "$push": push,
                "$set": set
            }
        }



import inspect
import typing
import types
import dataclasses
import enum

from .handle_args import get_input_dict, InputTyping, get_output_dict, Type

class Spec(enum.Enum):
    API_METHOD = '__ywpi_api_method__'
    CLASS_API_METHODS = '__ywpi_class_api_methods__'


@dataclasses.dataclass
class MethodDescription:
    parameters: list[inspect.Parameter]
    return_annotation: inspect.Parameter
    bind_method: bool = True


def service(cls):
    api_methods = {}
    for attrname in cls.__dict__:
        attrname: str
        if attrname.startswith('__'):
            continue

        if not hasattr(cls.__dict__[attrname], '__call__'):
            continue

        func = cls.__dict__[attrname]

        if hasattr(func, Spec.API_METHOD.value):
            signature = inspect.signature(func)

            paramenters = list(signature.parameters.values())
            bind_method = isinstance(func, types.FunctionType)
            if bind_method:
                paramenters.pop(0)

            api_methods[func.__name__] = MethodDescription(
                parameters=paramenters,
                return_annotation=signature.return_annotation,
                bind_method=bind_method
            )

    setattr(cls, Spec.CLASS_API_METHODS.value, api_methods)
    return cls


def api(func):
    setattr(func, Spec.API_METHOD.value, {})
    return func


@dataclasses.dataclass
class RegisteredMethod:
    fn: typing.Callable
    inputs: dict[str, InputTyping]
    outputs: dict[str, Type]
    description: typing.Optional[str] = None
    labels: typing.Optional[list[str]] = None


REGISTERED_METHODS: dict[str, RegisteredMethod] = {}
DEFAULT_DESCRIPTION = 'No description provided'


def _register_method(func, description: str = None, labels: list[str] = None):
    inputs = get_input_dict(func)
    outputs = get_output_dict(func)

    if description is None:
        description = func.__doc__

    REGISTERED_METHODS[func.__name__] = RegisteredMethod(
        fn=func,
        inputs=inputs,
        outputs=outputs,
        description=description,
        labels=labels
    )
    return func


def method(func=None, description: str = None, labels: list[str] = None):
    if func is None:
        def wrapper(func):
            return _register_method(
                func=func,
                description=description,
                labels=labels,
            )
        return wrapper
    else:
        return _register_method(
            func=func,
            description=description,
            labels=labels,
        )



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc
import warnings

from . import hub_pb2 as hub__pb2

GRPC_GENERATED_VERSION = '1.64.0'
GRPC_VERSION = grpc.__version__
EXPECTED_ERROR_RELEASE = '1.65.0'
SCHEDULED_RELEASE_DATE = 'June 25, 2024'
_version_not_supported = False

try:
    from grpc._utilities import first_version_is_lower
    _version_not_supported = first_version_is_lower(GRPC_VERSION, GRPC_GENERATED_VERSION)
except ImportError:
    _version_not_supported = True

if _version_not_supported:
    warnings.warn(
        f'The grpc package installed is at version {GRPC_VERSION},'
        + f' but the generated code in hub_pb2_grpc.py depends on'
        + f' grpcio>={GRPC_GENERATED_VERSION}.'
        + f' Please upgrade your grpc module to grpcio>={GRPC_GENERATED_VERSION}'
        + f' or downgrade your generated code using grpcio-tools<={GRPC_VERSION}.'
        + f' This warning will become an error in {EXPECTED_ERROR_RELEASE},'
        + f' scheduled for release on {SCHEDULED_RELEASE_DATE}.',
        RuntimeWarning
    )


class HubStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.Connect = channel.stream_stream(
                '/Hub/Connect',
                request_serializer=hub__pb2.Message.SerializeToString,
                response_deserializer=hub__pb2.Message.FromString,
                _registered_method=True)
        self.PushTask = channel.unary_unary(
                '/Hub/PushTask',
                request_serializer=hub__pb2.PushTaskRequest.SerializeToString,
                response_deserializer=hub__pb2.PushTaskResponse.FromString,
                _registered_method=True)
        self.RunTask = channel.unary_unary(
                '/Hub/RunTask',
                request_serializer=hub__pb2.PushTaskRequest.SerializeToString,
                response_deserializer=hub__pb2.RunTaskResponse.FromString,
                _registered_method=True)
        self.GetAgentsList = channel.unary_unary(
                '/Hub/GetAgentsList',
                request_serializer=hub__pb2.GetAgentsListRequest.SerializeToString,
                response_deserializer=hub__pb2.GetAgentsListResponse.FromString,
                _registered_method=True)
        self.SubscribeOnAgents = channel.unary_stream(
                '/Hub/SubscribeOnAgents',
                request_serializer=hub__pb2.SubscribeOnAgentsRequest.SerializeToString,
                response_deserializer=hub__pb2.SubscribeOnAgentsResponse.FromString,
                _registered_method=True)


class HubServicer(object):
    """Missing associated documentation comment in .proto file."""

    def Connect(self, request_iterator, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def PushTask(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def RunTask(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetAgentsList(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SubscribeOnAgents(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_HubServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'Connect': grpc.stream_stream_rpc_method_handler(
                    servicer.Connect,
                    request_deserializer=hub__pb2.Message.FromString,
                    response_serializer=hub__pb2.Message.SerializeToString,
            ),
            'PushTask': grpc.unary_unary_rpc_method_handler(
                    servicer.PushTask,
                    request_deserializer=hub__pb2.PushTaskRequest.FromString,
                    response_serializer=hub__pb2.PushTaskResponse.SerializeToString,
            ),
            'RunTask': grpc.unary_unary_rpc_method_handler(
                    servicer.RunTask,
                    request_deserializer=hub__pb2.PushTaskRequest.FromString,
                    response_serializer=hub__pb2.RunTaskResponse.SerializeToString,
            ),
            'GetAgentsList': grpc.unary_unary_rpc_method_handler(
                    servicer.GetAgentsList,
                    request_deserializer=hub__pb2.GetAgentsListRequest.FromString,
                    response_serializer=hub__pb2.GetAgentsListResponse.SerializeToString,
            ),
            'SubscribeOnAgents': grpc.unary_stream_rpc_method_handler(
                    servicer.SubscribeOnAgents,
                    request_deserializer=hub__pb2.SubscribeOnAgentsRequest.FromString,
                    response_serializer=hub__pb2.SubscribeOnAgentsResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'Hub', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('Hub', rpc_method_handlers)


 # This class is part of an EXPERIMENTAL API.
class Hub(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def Connect(request_iterator,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.stream_stream(
            request_iterator,
            target,
            '/Hub/Connect',
            hub__pb2.Message.SerializeToString,
            hub__pb2.Message.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def PushTask(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/Hub/PushTask',
            hub__pb2.PushTaskRequest.SerializeToString,
            hub__pb2.PushTaskResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def RunTask(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/Hub/RunTask',
            hub__pb2.PushTaskRequest.SerializeToString,
            hub__pb2.RunTaskResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetAgentsList(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/Hub/GetAgentsList',
            hub__pb2.GetAgentsListRequest.SerializeToString,
            hub__pb2.GetAgentsListResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def SubscribeOnAgents(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(
            request,
            target,
            '/Hub/SubscribeOnAgents',
            hub__pb2.SubscribeOnAgentsRequest.SerializeToString,
            hub__pb2.SubscribeOnAgentsResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)



import inspect
import typing as t
import dataclasses
import pydantic

from ywpi.handle_args import SERIALIZERS, DESERIALIZERS, TYPE_NAMES

def serial(fn):
    """
    Build referenced json from python object.
    Method should return json serializable dict.

    Example of serializer that serialize CustomType to dict
    def fn(data: CustomType) -> dict: pass
    """
    data_tp = inspect.signature(fn).return_annotation
    assert data_tp in (str, int, float, dict, list)

    params = list(inspect.signature(fn).parameters.values())
    assert len(params) == 1
    tp = params[0].annotation
    assert not issubclass(inspect.Parameter.empty, tp)
    assert tp not in SERIALIZERS
    SERIALIZERS[tp] = fn
    return fn


def deserial(fn):
    """
    Build some python object from part of ywpi method input.
    Input could be one of allowed JSON filelds types (str, int, float, dict, list)

    Example of deserializer that require dict as input
    def fn(data: dict) -> CustomType: pass

    Example of deserializer that specify type name
    def fn(data: dict) -> t.Annotated[CustomType, "custom_type_name"]: pass
    """
    tp = inspect.signature(fn).return_annotation

    if t.get_origin(tp) is t.Annotated:
        tp_name = tp.__metadata__[0]
        tp = tp.__origin__
        assert type(tp_name) is str
    else:
        tp_name = tp.__name__

    # TODO: Remove assert and use exceptions
    assert not issubclass(inspect.Parameter.empty, tp)
    assert tp not in DESERIALIZERS

    # TODO: Handle input data type
    params = list(inspect.signature(fn).parameters.values())
    assert len(params) == 2
    data_tp = params[0].annotation
    assert not issubclass(inspect.Parameter.empty, data_tp)
    assert data_tp in (str, int, float, dict, list)

    DESERIALIZERS[tp] = fn
    TYPE_NAMES[tp] = tp_name
    return fn


def handle_outputs(data: t.Any) -> dict[str, t.Any]:
    if isinstance(data, dict):
        return data
    elif isinstance(data, (list, tuple, set)):
        return {
            '__others__': [ handle_outputs(d) for d in data ]
        }
    elif isinstance(data, pydantic.BaseModel):
        return data.model_dump(mode='json')
    elif dataclasses.is_dataclass(data):
        return dataclasses.asdict(data)
    elif type(data) in SERIALIZERS:
        return SERIALIZERS[type(data)](data)
    else:
        raise TypeError(f'Type {type(data)} has not got serialize handler')


# @dataclasses.dataclass
# class Ref:
#     ref: str
#     type: str

# # To ywpi type (serialization)
# @serializer
# def type_cvt(value: int):
#     if not isinstance(value, int):
#         raise TypeError(value)
#     return value

# # From ywpi type (deserialization)
# @deserializer
# def cvt(value) -> int:
#     # Downlaod image by ref
#     return value


def validate_bytes_with_attachments_ctx(data, info: pydantic.ValidationInfo) -> bytes:
    if isinstance(data, dict):
        try:
            attachments = info.context["attachments"]
        except:
            raise ValueError(f'attachments does not present in the context')

        try:
            key = data["$attachment"]
        except:
            raise ValueError('data does not has "$attachment" attribute')

        try:
            return attachments[key]
        except:
            raise ValueError(f'attachment with key "{key}" does not present in the context')

    return data


BytesValidator = pydantic.BeforeValidator(validate_bytes_with_attachments_ctx)
"""
Serializer require context with *attachments* dictionary.
"""


def serialize_bytes_with_attachments_ctx(data: bytes, info: pydantic.SerializationInfo):
    info.context["attachments"]["1"] = data
    return {
        "$attachment": "1"
    }


BytesSerializer = pydantic.PlainSerializer(serialize_bytes_with_attachments_ctx)
"""
Serializer require context with *attachments* dictionary.
Serializer replace binary content with reference and insert data into *attachments* dictionary,
"""



def _track(fn):
    def wrapper():
        fn()

    return wrapper


def track(fn=None, track_yields=True):
    if fn is None:
        return _track(fn)

    def decorator(fn):
        return _track(fn)
    
    return decorator



@track()
def some_trackable(number: int, word: str):
    yield number
    yield word

    return f'{word}-{number}'







import importlib
import argparse
import sys
import os
from pathlib import Path

import watchfiles

import ywpi
from ywpi.logger import logger
import ywpi.stub
import ywpi.stub.generation

parser = argparse.ArgumentParser()
command_subparsers = parser.add_subparsers(dest='command')

run_parser = command_subparsers.add_parser('run')
run_parser.add_argument('target', help='Path to launch target. Example: agents.test:func')
run_parser.add_argument('--id', type=str, help='Agent id', default='ywpi-run')
run_parser.add_argument('--name', type=str, help='Agent name', default='Untitled')
run_parser.add_argument('--project', type=str, help='Project ID')
run_parser.add_argument('--reload', action='store_true', help='Enable auto reloading', default=False)
run_parser.add_argument('--auto-reconnect', action='store_true', help='Enable auto reconnection to hub', default=False)

command_subparsers.add_parser('methods')

sync_parser = command_subparsers.add_parser('sync')
sync_parser.add_argument('stub_file_path', help='Path to file with stub. Example: app/stub.py')


def perform_run_command(args):
    module_path, _, method = args.target.partition(':')

    sys.path.insert(0, os.getcwd())

    # Truncate `.py` extension
    # and replace `/` with `.` for fs path handling (i.e. `agets/search.py`)
    if module_path.endswith('.py'):
        module_path = module_path[:-3]
    module_path = module_path.replace('/', '.')

    module = importlib.import_module(module_path)

    # Register method
    if len(method) > 0:
        ywpi.method(module.__dict__[method])

    kwargs = {}
    if args.project is not None: kwargs['project'] = args.project

    if not args.auto_reconnect:
        ywpi.serve(args.id, args.name, **kwargs)
    else:
        ywpi.serve_with_reconnecting(args.id, args.name, **kwargs)


def perform_sync(args):
    stub_file_path = Path(args.stub_file_path)
    pyi_file_path = stub_file_path.with_suffix(".pyi")

    methods = ywpi.get_methods()
    content = ywpi.stub.generation.generate_stub_file_content(methods)
    
    with open(pyi_file_path, 'w') as file:
        file.write(content)


def main():
    args = parser.parse_args()

    def callback(e):
        path = os.path.relpath(next(iter(e))[1], os.getcwd())
        logger.warning(f"WatchFiles detected changes in '{path}'. Reloading...")

    if args.command == 'run':
        if args.reload:
            watchfiles.run_process('', target=perform_run_command, args=(args,), callback=callback)
        else:
            perform_run_command(args)
    elif args.command == 'methods':
        for m in ywpi.get_methods():
            print(f'{m.agent_id}/{m.name}')
    elif args.command == 'sync':
        perform_sync(args)


if __name__ == '__main__':
    main()



import logging

FORMATTER = logging.Formatter("[%(asctime)s] %(levelname)s %(message)s")

logger = logging.getLogger('app')
# logger.setLevel(logging.DEBUG)
logger.setLevel(logging.INFO)

console_handler = logging.StreamHandler()
logger.addHandler(console_handler)
console_handler.setFormatter(FORMATTER)



import typing as t
import functools
from inspect import signature

import pydantic

import ywpi


# TODO: # Add support for LexicalNode, list[LexicalNode], Markdown
CompletitionFnT = t.Callable[[], t.Union[str]]


class CompletitionOutputs(pydantic.BaseModel):
    completition: t.Optional[str] = None


def completition(fn: CompletitionFnT):
    """
    Decorate function that implement "autocomplete" (completition) workflow

    Possible inputs:
    - Context[Markdown]
    - Selection
    """

    @functools.wraps(fn)
    def wrapper(*args, **kwargs):
        return CompletitionOutputs(completition=fn(*args, **kwargs))

    # Replace outputs annotation
    sig = signature(wrapper)
    wrapper.__signature__ = sig.replace(
        parameters=list(sig.parameters.values()),
        return_annotation=CompletitionOutputs
    )

    return ywpi.method(wrapper, labels=["contexts/completition"])



from concurrent import futures

import pydantic

from . import hub_pb2
from . import hub_pb2_grpc
from . import hub_models


class AbstractConnection:
    def call(
        self,
        rpc: hub_pb2.Rpc,
        payload: pydantic.BaseModel
    ) -> futures.Future[hub_pb2.ResponseMessage]: raise NotImplementedError()

    def call_register_agent(
        self,
        payload: hub_models.RegisterAgentRequest
    ) -> futures.Future[hub_pb2.ResponseMessage]: raise NotImplementedError()

    def call_update_task(
        self,
        payload: hub_models.UpdateTaskRequest
    ) -> futures.Future[hub_pb2.ResponseMessage]: raise NotImplementedError()



import typing as t
from functools import wraps
from inspect import signature, isgeneratorfunction

import pydantic
import ywpi


T = t.TypeVar('T')


class Subscribtion(pydantic.BaseModel):
    agent_id: str
    method_name: str


class ContextUpdate(pydantic.BaseModel):
    context_id: str
    update: dict

    @pydantic.model_serializer()
    def serialize_model(self):
        return {
            "type": "ContextUpdate",
            "data": {
                "context_id": self.context_id,
                "update": self.update
            }
        }


class Context(pydantic.BaseModel, t.Generic[T]):
    id: str
    project_id: t.Optional[str] = None
    labels: t.Optional[list[ywpi.Label]] = None    
    subscribtions: list[Subscribtion]
    tp: str # type: str
    data: T

    def build_update(self, push={}, set={}):
        return ContextUpdate(
            context_id=self.id,
            update={
                "$push": push,
                "$set": set
            }
        )


class ContextId(pydantic.BaseModel):
    id: str


class InMemoryState(dict): ...


IN_MEMORY_STATES: dict[str, InMemoryState] = {}


class Chat(pydantic.BaseModel):
    messages: list[dict] = []
    tools: list = []


def agent(fn):
    """
    Wrapper for chat agents
    ----
    
    Allow to recieve `State` parameter that statefull in-memory storage binded to context ID.

    Automatically handle `str` outputs and convert that to assistant message.
    """
    
    source_signature = signature(fn)
    parameters = []

    state_param_name = None
    for name, parameter in source_signature.parameters.items():
        if parameter.annotation is InMemoryState:
            state_param_name = name
        else:
            parameters.append(parameter)

    @wraps(fn)
    def wrapper(*args, **kwargs):
        context = None
        for k, v in kwargs.items():
            if isinstance(v, Context):
                context = v

        if context is None:
            raise RuntimeError("Context type not found in inputs")

        if not isgeneratorfunction(fn):
            if state_param_name is not None:
                if context.id not in IN_MEMORY_STATES:
                    IN_MEMORY_STATES[context.id] = InMemoryState()

                kwargs[state_param_name] = IN_MEMORY_STATES[context.id]

            outputs = fn(*args, **kwargs)

            if isinstance(outputs, str):
                yield context.build_update(push={
                    "messages": {
                        "role": "assistant",
                        "content": outputs
                    }
                })
            else:
                yield outputs if outputs is not None else {}
        else:
            if state_param_name is not None:
                if context.id not in IN_MEMORY_STATES:
                    IN_MEMORY_STATES[context.id] = InMemoryState()

                kwargs[state_param_name] = IN_MEMORY_STATES[context.id]

            for outputs in fn(*args, **kwargs):
                if outputs is None:
                    yield {}

                if isinstance(outputs, str):
                    yield context.build_update(push={
                        "messages": {
                            "role": "assistant",
                            "content": outputs
                        }
                    })
                else:
                    yield outputs if outputs is not None else {}

    wrapper.__signature__ = source_signature.replace(parameters=parameters)
    ywpi.decorators._register_method(wrapper)

    return wrapper


class LexicalNode(pydantic.BaseModel):
    children: list['LexicalNode'] = []
    type: str
    direction: t.Optional[t.Any] = None
    format: t.Any = None
    indent: int = 0
    textFormat: t.Optional[int] = 0
    textStyle: str = ""
    uuid: str = ""

    text: t.Optional[str] = ""

    _key: str = None
    _index: int = None
    _parent: t.Optional['LexicalNode'] = None
    _virtual: bool = False

    def insert_after(self, node: 'LexicalNode') -> dict:
        """Return update dict"""
        return {
            "$push": {
                f"{self._parent._key}.children": {
                    "$each": [node],
                    "$position": self._index + 1
                }
            }
        }

    def insert_before(self, node: 'LexicalNode') -> dict:
        """Return update dict"""
        return {
            "$push": {
                f"{self._parent._key}.children": {
                    "$each": [node],
                    "$position": self._index
                }
            }
        }

    def append(self, node: 'LexicalNode') -> dict:
        """Return update dict"""
        return {
            "$push": {
                f"{self._parent._key}.children": node
            }
        }

    @property
    def content(self) -> str:
        if self.type == "text":
            return self.text
        elif self.type == "paragraph":
            return "".join(
                map(lambda e: e.content, self.children)
            )
        elif self.type == "root":
            return "\n\n".join(
                map(lambda e: e.content, self.children)
            )
        else:
            return self.text

    def content_left(self, key: 'Selection.Anchor') -> str:
        if self.type == "text":
            return self.text, False
        elif self.type == "paragraph":
            res = ""
            for child in self.children:
                child_res, flag = child.content_left(key)

                if child._key == key.key:
                    res += child_res[:key.offset]
                    return res, True
                else:
                    res += child_res

                if flag == True:
                    return res, flag

            return res, False
        elif self.type == "root":
            res = ""
            for child in self.children:
                if child._key == key.key:
                    break

                child_res, flag = child.content_left(key)
                res += "\n\n" + child_res

                if flag == True:
                    return res, flag
            return res, False
        else:
            return self.text, False



class EditorState(pydantic.BaseModel):
    root: LexicalNode

    @staticmethod
    def __update_tree(node: LexicalNode, path: str, index: int | None = None, parent: LexicalNode | None = None):
        for idx, child_node in enumerate(node.children):
            EditorState.__update_tree(child_node, f"{path}.children.{idx}", index=idx, parent=node)

        node._key = path
        if index is not None:
            node._index = index
        if parent is not None:
            node._parent = parent

    @pydantic.model_validator(mode='after')
    def _update_internals(self):
        """
        Add additional fields `_key`, `_index`, `_parent` to build a nodes tree.
        Fields required for simplifying tree update.
        Internal `_key` not equal or linked with lexical `key`.
        """
        self.__update_tree(self.root, "tree.root")
        return self


class Markdown(pydantic.BaseModel):
    content: str = None
    tree: EditorState = None


def append_keys(state: dict, path: str, index: int = None):
    if "children" in state:
        for idx, c in enumerate(state["children"]):
            append_keys(c, f"{path}.children.{idx}", index=idx)

    state["_key"] = path
    if index is not None:
        state["_index"] = index


class Selection(pydantic.BaseModel):
    class Anchor(pydantic.BaseModel):
        key: str
        type: str
        offset: t.Optional[int] = None

    nodes: dict[str, LexicalNode]
    text: str
    anchor: 'Selection.Anchor'

    @pydantic.model_validator(mode='after')
    def _update_internals(self):
        """
        Add additional fields `_key`, `_index`, `_parent` to build a nodes tree.
        Fields required for simplifying tree update.
        Internal `_key` not equal or linked with lexical `key`.
        """
        for k, v in self.nodes.items():
            v._key = k
            v._index = int(k.split(".")[-1])
            indexes = tuple(filter(lambda e: e.isdigit(), k.split(".")))
            parent_index = indexes[:-1]
            parent = "tree.root." + ".".join(map(lambda e: f"children.{e}", parent_index))

            if parent in self.nodes: # Parent also in the selection
                v._parent = self.nodes[parent]
            else:
                key = parent.removeprefix("tree.root.")
                if not len(key):
                    # TODO: Reuse virtual node
                    root = LexicalNode(type="root")
                    root._key = "tree.root"
                    root._index = 0
                    root._virtual = True
                    v._parent = root
                else:
                    virtual_parent_node = LexicalNode(type="undefined")
                    virtual_parent_node._key = "tree.root." + key
                    virtual_parent_node._index = int(k.split(".")[-1])
                    virtual_parent_node._virtual = True
                    v._parent = virtual_parent_node
                    # We know full paths and all parents (without additional info)
                    # But has not got any
                    # TODO: Handle this case
                    # print("Warning: required virtual parent for", parent, key)

        return self

    @pydantic.model_validator(mode='after')
    def _update_internals(self):
        """
        Add additional fields `_key`, `_index`, `_parent` to build a nodes tree.
        Fields required for simplifying tree update.
        Internal `_key` not equal or linked with lexical `key`.
        """
        for k, v in self.nodes.items():
            v._key = k
            v._index = int(k.split(".")[-1])
            indexes = tuple(filter(lambda e: e.isdigit(), k.split(".")))
            parent_index = indexes[:-1]
            parent = "tree.root." + ".".join(map(lambda e: f"children.{e}", parent_index))

            if parent in self.nodes: # Parent also in the selection
                v._parent = self.nodes[parent]
            else:
                key = parent.removeprefix("tree.root.")
                if not len(key):
                    # TODO: Reuse virtual node
                    root = LexicalNode(type="root")
                    root._key = "tree.root"
                    root._index = 0
                    root._virtual = True
                    v._parent = root
                else:
                    virtual_parent_node = LexicalNode(type="undefined")
                    virtual_parent_node._key = "tree.root." + key
                    virtual_parent_node._index = int(k.split(".")[-1])
                    virtual_parent_node._virtual = True
                    v._parent = virtual_parent_node
                    # We know full paths and all parents (without additional info)
                    # But has not got any
                    # TODO: Handle this case
                    # print("Warning: required virtual parent for", parent, key)

        return self




# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: hub.proto
# Protobuf Python Version: 5.26.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\thub.proto\"4\n\x0eRequestMessage\x12\x11\n\x03rpc\x18\x01 \x01(\x0e\x32\x04.Rpc\x12\x0f\n\x07payload\x18\x03 \x01(\t\"Q\n\x0fResponseMessage\x12\x14\n\x07payload\x18\x02 \x01(\tH\x00\x88\x01\x01\x12\x12\n\x05\x65rror\x18\x03 \x01(\tH\x01\x88\x01\x01\x42\n\n\x08_payloadB\x08\n\x06_error\"\xd4\x01\n\x07Message\x12\x10\n\x08reply_to\x18\x01 \x01(\t\x12\"\n\x07request\x18\x02 \x01(\x0b\x32\x0f.RequestMessageH\x00\x12$\n\x08response\x18\x03 \x01(\x0b\x32\x10.ResponseMessageH\x00\x12.\n\x0b\x61ttachments\x18\x04 \x03(\x0b\x32\x19.Message.AttachmentsEntry\x1a\x32\n\x10\x41ttachmentsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x0c:\x02\x38\x01\x42\t\n\x07message\"W\n\x05\x45rror\x12\x0c\n\x04type\x18\x01 \x01(\t\x12\x16\n\ttraceback\x18\x02 \x01(\tH\x00\x88\x01\x01\x12\x11\n\x04\x64\x61ta\x18\x03 \x01(\tH\x01\x88\x01\x01\x42\x0c\n\n_tracebackB\x07\n\x05_data\"t\n\x0fPushTaskRequest\x12\x10\n\x08\x61gent_id\x18\x01 \x01(\t\x12\x0e\n\x06method\x18\x02 \x01(\t\x12\x0e\n\x06params\x18\x03 \x01(\t\x12\x0f\n\x07payload\x18\x04 \x01(\t\x12\x13\n\x06silent\x18\x05 \x01(\x08H\x00\x88\x01\x01\x42\t\n\x07_silent\"Z\n\x10PushTaskResponse\x12\x14\n\x07task_id\x18\x01 \x01(\tH\x00\x88\x01\x01\x12\x1a\n\x05\x65rror\x18\x02 \x01(\x0b\x32\x06.ErrorH\x01\x88\x01\x01\x42\n\n\x08_task_idB\x08\n\x06_error\"{\n\x0fRunTaskResponse\x12\x14\n\x07task_id\x18\x01 \x01(\tH\x00\x88\x01\x01\x12\x1a\n\x05\x65rror\x18\x02 \x01(\x0b\x32\x06.ErrorH\x01\x88\x01\x01\x12\x14\n\x07outputs\x18\x03 \x01(\tH\x02\x88\x01\x01\x42\n\n\x08_task_idB\x08\n\x06_errorB\n\n\x08_outputs\"\x16\n\x14GetAgentsListRequest\"#\n\x05Input\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0c\n\x04type\x18\x02 \x01(\t\".\n\x06Method\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x16\n\x06inputs\x18\x02 \x03(\x0b\x32\x06.Input\";\n\x05\x41gent\x12\n\n\x02id\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x18\n\x07methods\x18\x03 \x03(\x0b\x32\x07.Method\"/\n\x15GetAgentsListResponse\x12\x16\n\x06\x61gents\x18\x01 \x03(\x0b\x32\x06.Agent\"\x1a\n\x18SubscribeOnAgentsRequest\"c\n\x19SubscribeOnAgentsResponse\x12\x14\n\x07payload\x18\x01 \x01(\tH\x00\x88\x01\x01\x12\x1a\n\x05\x65rror\x18\x02 \x01(\x0b\x32\x06.ErrorH\x01\x88\x01\x01\x42\n\n\x08_payloadB\x08\n\x06_error*\x86\x01\n\x03Rpc\x12\x11\n\rRPC_UNDEFINED\x10\x00\x12\x16\n\x12RPC_REGISTER_AGENT\x10\x01\x12\x17\n\x13RPC_HEARTBEAT_AGENT\x10\x02\x12\x12\n\x0eRPC_START_TASK\x10\x03\x12\x12\n\x0eRPC_ABORT_TASK\x10\x04\x12\x13\n\x0fRPC_UPDATE_TASK\x10\x05\x32\x96\x02\n\x03Hub\x12!\n\x07\x43onnect\x12\x08.Message\x1a\x08.Message(\x01\x30\x01\x12/\n\x08PushTask\x12\x10.PushTaskRequest\x1a\x11.PushTaskResponse\x12-\n\x07RunTask\x12\x10.PushTaskRequest\x1a\x10.RunTaskResponse\x12>\n\rGetAgentsList\x12\x15.GetAgentsListRequest\x1a\x16.GetAgentsListResponse\x12L\n\x11SubscribeOnAgents\x12\x19.SubscribeOnAgentsRequest\x1a\x1a.SubscribeOnAgentsResponse0\x01\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'hub_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_MESSAGE_ATTACHMENTSENTRY']._loaded_options = None
  _globals['_MESSAGE_ATTACHMENTSENTRY']._serialized_options = b'8\001'
  _globals['_RPC']._serialized_start=1138
  _globals['_RPC']._serialized_end=1272
  _globals['_REQUESTMESSAGE']._serialized_start=13
  _globals['_REQUESTMESSAGE']._serialized_end=65
  _globals['_RESPONSEMESSAGE']._serialized_start=67
  _globals['_RESPONSEMESSAGE']._serialized_end=148
  _globals['_MESSAGE']._serialized_start=151
  _globals['_MESSAGE']._serialized_end=363
  _globals['_MESSAGE_ATTACHMENTSENTRY']._serialized_start=302
  _globals['_MESSAGE_ATTACHMENTSENTRY']._serialized_end=352
  _globals['_ERROR']._serialized_start=365
  _globals['_ERROR']._serialized_end=452
  _globals['_PUSHTASKREQUEST']._serialized_start=454
  _globals['_PUSHTASKREQUEST']._serialized_end=570
  _globals['_PUSHTASKRESPONSE']._serialized_start=572
  _globals['_PUSHTASKRESPONSE']._serialized_end=662
  _globals['_RUNTASKRESPONSE']._serialized_start=664
  _globals['_RUNTASKRESPONSE']._serialized_end=787
  _globals['_GETAGENTSLISTREQUEST']._serialized_start=789
  _globals['_GETAGENTSLISTREQUEST']._serialized_end=811
  _globals['_INPUT']._serialized_start=813
  _globals['_INPUT']._serialized_end=848
  _globals['_METHOD']._serialized_start=850
  _globals['_METHOD']._serialized_end=896
  _globals['_AGENT']._serialized_start=898
  _globals['_AGENT']._serialized_end=957
  _globals['_GETAGENTSLISTRESPONSE']._serialized_start=959
  _globals['_GETAGENTSLISTRESPONSE']._serialized_end=1006
  _globals['_SUBSCRIBEONAGENTSREQUEST']._serialized_start=1008
  _globals['_SUBSCRIBEONAGENTSREQUEST']._serialized_end=1034
  _globals['_SUBSCRIBEONAGENTSRESPONSE']._serialized_start=1036
  _globals['_SUBSCRIBEONAGENTSRESPONSE']._serialized_end=1135
  _globals['_HUB']._serialized_start=1275
  _globals['_HUB']._serialized_end=1553
# @@protoc_insertion_point(module_scope)



import json

import grpc
import pydantic

from ywpi import settings
from ywpi import hub_pb2
from ywpi import hub_pb2_grpc
from ywpi import hub_models

grpc_channel = grpc.insecure_channel(settings.YWPI_HUB_HOST)
hub_stub = hub_pb2_grpc.HubStub(grpc_channel)


class Method:
    def __init__(self, agent_id: str, name: str, inputs: list[hub_pb2.Input] = []):
        self.agent_id = agent_id
        self.name = name
        self.inputs = inputs

    def __repr__(self):
        return f'Method(agent={self.agent_id}, name={self.name}, inputs={self.inputs})'

    def _create_request(self, *args, **kwargs):
        if len(args):
            raise RuntimeError('ywpi method can not recieve positional args')

        params = {
            key: value.model_dump(mode='json') if isinstance(value, pydantic.BaseModel) else value
            for key, value in kwargs.items()
        }

        return hub_pb2.PushTaskRequest(
            agent_id=self.agent_id,
            method=self.name,
            params=json.dumps(params)
        )

    def _handle_response(self, response: hub_pb2.RunTaskResponse):
        if response.HasField('error'):
            error = response.error
            data = json.loads(error.data) if error.HasField("data") else None
            traceback = error.traceback
            exception = hub_models.YwpiException.from_data(type=error.type, data=data, traceback=traceback)
            raise exception

        outputs = json.loads(response.outputs)
        if '__others__' in outputs:
            return outputs['__others__']
        return outputs

    def __call__(self, *args, **kwds):
        return self._handle_response(
            hub_stub.RunTask(
                self._create_request(*args, **kwds)
            )
        )

    def to_autogen_tool(self):
        pass

    # def to_langchain_tool(other_self):

    #     class CustomTool(BaseTool):
    #         def _run(self, *args, **kwargs):
    #             return other_self.__call__(*args, **kwargs)

    #     return CustomTool(name=other_self.name, description='', args_schema={
    #         "definitions": {
    #             "value": {
    #                 "type": "integer"
    #             }
    #         }
    #     })


def get_methods() -> list[Method]:
    result = []
    response: hub_pb2.GetAgentsListResponse = hub_stub.GetAgentsList(hub_pb2.GetAgentsListRequest())

    for agent in response.agents:
        for method in agent.methods:
            result.append(Method(
                agent.id,
                method.name,
                method.inputs
            ))
    
    return result


def get_method(agent: str, name: str) -> Method:
    response: hub_pb2.GetAgentsListResponse = hub_stub.GetAgentsList(hub_pb2.GetAgentsListRequest())

    for a in response.agents:
        if a.id != agent:
            continue

        for m in a.methods:
            if m.name != name:
                continue
            return Method(
                a.id,
                m.name,
                m.inputs
            )
    raise RuntimeError('Method not found')



import typing as t
import io

import pydantic
try:
    import requests
except:
    pass

from ywpi.handle_args import TYPE_CONVERTERS, DESERIALIZERS, TYPE_NAMES


try:
    import pymupdf

    def to_pymupdf_document(pdf: 'PDF') -> pymupdf.Document:
        return pymupdf.Document(stream=pdf._download_file())

    TYPE_CONVERTERS[('PDF', pymupdf.Document)] = to_pymupdf_document
except:
    pass


class PDF(pydantic.BaseModel):
    name: str
    src: str

    @staticmethod
    def from_data(data: dict, *args) -> 'PDF':
        return PDF.model_validate(data)

    def _download_file(self) -> bytes:
        print(f'Downloading from {self.src}')
        res = requests.get(self.src)
        res.raise_for_status()
        print(f'Downloading complete {self.src}')
        return res.content

    @staticmethod
    def to_bytes(pdf: 'PDF'):
        return b'File-Content'

    @staticmethod
    def to_str(pdf: 'PDF'):
        file_content = pdf._download_file()
        print('Extracting text')
        with io.BytesIO(file_content) as file:
            doc = pymupdf.open(stream=file)
            return ' '.join(map(lambda e: e.get_text(), doc))


TYPE_NAMES[PDF] = 'pdf'
DESERIALIZERS[PDF] = PDF.from_data
TYPE_CONVERTERS[(PDF, bytes)] = PDF.to_bytes
TYPE_CONVERTERS[(PDF, str)] = PDF.to_str


class DocumentText(pydantic.BaseModel):
    page_number: int
    text: str
    x1: float
    y1: float
    x2: float
    y2: float

    @staticmethod
    def from_pymupdf_tuple(w, page_number: int):
        return DocumentText(
            page_number=page_number,
            text=w[4],
            x1=w[0],
            y1=w[1],
            x2=w[2],
            y2=w[3]
        )

TYPE_NAMES[DocumentText] = 'document_text'
DESERIALIZERS[PDF] = PDF.from_data


class Message(pydantic.BaseModel):
    id: t.Optional[str] = None
    role: str
    content: str
TYPE_NAMES[Message] = 'message'


class Chat(pydantic.BaseModel):
    messages: list[Message]
TYPE_NAMES[Chat] = 'chat'


__all__ = (
    'PDF',
    'DocumentText',
    'Message',
    'Chat'
)



import typing as t
import threading
import traceback

from .connection import AbstractConnection
from . import serialization
from . import hub_models
from .logger import logger
from .handle_args import handle_args as params_from_dict
from .stream import Stream

T = t.TypeVar('T')


class Queue(t.Generic[T]):
    """
    Implement custom queue because of requiring `shutdown` method.
    Only Python 3.13 Queue has `shutdown` method.
    """
    def __init__(self, init_items: list | None = None):
        self._lock = threading.Lock()
        self._cond = threading.Condition(self._lock)
        self._items: list[T] = []
        if init_items is not None:
            self._items = init_items.copy()
        self._terminated = False

    def put(self, item: T):
        with self._lock:
            self._items.append(item)

    def terminate(self):
        with self._cond:
            self._terminated = True
            self._cond.notify_all()

    def get(self) -> T:
        with self._cond:
            if len(self._items) <= 0 and not self._terminated: # List empty
                self._cond.wait_for(lambda: len(self._items) > 0 or self._terminated)

                if len(self._items) <= 0 or self._terminated:
                    raise StopIteration()

            return self._items.pop(0)


class IOManager:
    def __init__(
        self,
        task_id: str,
        io_schema: dict, # Actually inputs schema
        connection: AbstractConnection
    ):
        self._lock = threading.Lock()
        self._cv = threading.Condition()
        self._streams: dict[str, Queue] = {}
        self._connection = connection
        self._task_id = task_id
        self._io_schema = io_schema

    @property
    def task_id(self):
        return self._task_id

    def handle_outputs(self, outputs: t.Union[dict, tuple, t.Any]):
        """
        Handle method outputs.
        """
        return self._connection.call_update_task(
            hub_models.UpdateTaskRequest(id=self._task_id, outputs=serialization.handle_outputs(outputs))
        )

    def handle_inputs(self, inputs: dict[str, t.Any], attachments: t.MutableMapping[str, dict]) -> dict[str, t.Any]:
        params = params_from_dict(inputs, self._io_schema, attachments)

        streams: dict[str, Stream] = {}

        for k, v in params.items():
            if isinstance(v, Stream):
                streams[k] = v

        return params

    def update_task_status(self, status: str, final_outputs: dict = None, error: hub_models.Error = None):
        outputs = None
        try:
            outputs = serialization.handle_outputs(final_outputs) if final_outputs is not None else None
        except:
            logger.warning(f'Error while converting final task outputs: {traceback.format_exc()}')
            status = 'failed'

        return self._connection.call_update_task(hub_models.UpdateTaskRequest(
            id=self._task_id,
            status=status,
            outputs=outputs,
            error=error
        ))

    def append_stream(self, stream_id: str, init_items: list | None = None):
        with self._lock:
            if stream_id in self._streams:
                raise KeyError(f'stream "{stream_id}" already exists')
            self._streams[stream_id] = Queue(init_items)

    def get_stream_item(self, stream_id: str):
        """
        Stream call this method to retrieve next item.
        """
        with self._lock:
            if stream_id in self._streams:
                stream = self._streams[stream_id]
            else:
                raise KeyError()

        return stream.get()

    def update_inputs(self, inputs: dict[str, t.Any]):
        """
        Exchanger call this method to append item to stream.
        """
        with self._lock:
            for k, v in inputs:
                if k in self._streams:
                    self._streams[k].put(v)
                else:
                    logger.warning(f'Task "{self._task_id}" does not has stream "{k}"')



import warnings
warnings.simplefilter("ignore", UserWarning)

import threading
from concurrent import futures
import inspect
import typing
import uuid
import inspect
import traceback

import grpc
import pydantic

from . import hub_pb2
from . import hub_pb2_grpc
from . import hub_models
from .logger import logger
from . import settings
from ywpi import RegisteredMethod, REGISTERED_METHODS
from ywpi.handle_args import InputTyping
from .io_manager import IOManager
from .method_schemes import get_function_schema
from ywpi.exceptions import get_exception_serialization_model


class Channel:
    def __init__(self):
        self.lock = threading.Lock()
        self.condition = threading.Condition(lock=self.lock)
        self.messages = []
        self.running = True

    def push(self, message):
        with self.condition:
            self.messages.append(message)
            self.condition.notify_all()

    def close(self):
        self.running = False
        with self.condition:
            self.condition.notify_all()

    def __iter__(self):
        return self

    def __next__(self):
        with self.condition:
            while not self.messages and self.running:
                self.condition.wait()
            if not self.running:
                raise StopIteration()
            return self.messages.pop(0)


# Service level
# class ServiceServer:
#     def __init__(self, agent_cls, exchanger: 'Exchanger' = None) -> None:
#         self.agent_cls = agent_cls
#         self.thread_pool = futures.ThreadPoolExecutor(max_workers=1)

#         self.methods: list[models.Method] = []
#         self.calls: dict[str, typing.Callable] = {}
#         self.exchanger = exchanger

#         for name, description in self.agent_cls.__dict__[Spec.CLASS_API_METHODS.value].items():
#             description: MethodDescription
#             self.methods.append(hub_models.Method(
#                 name=name,
#                 inputs=[
#                     hub_models.InputDescription(name=param.name, type=ServiceServer.TYPE_TO_YWPI[param.annotation])
#                     for param in description.parameters
#                 ]
#             ))
#             self.calls[name] = self.agent_cls.__dict__[name]

#     @staticmethod
#     def _method_wrapper(task_id: str, exchanger: 'Exchanger', method, **kwargs):
#         try:
#             staticgenerator = isinstance(method, staticmethod) and inspect.isgeneratorfunction(method.__func__)
#             if inspect.isgeneratorfunction(method) or staticgenerator:
#                 for outputs in method(**kwargs):
#                     exchanger.call_update_task(hub_models.UpdateTaskRequest(id=task_id, outputs=outputs))
#             else:
#                 method(**kwargs)
#                 status = 'completed'
#         except BaseException as e:
#             import traceback
#             print(traceback.format_exc())
#             logger.warning('method raise exception')
#             status = 'failed'
#         finally:
#             exchanger.call_update_task(hub_models.UpdateTaskRequest(id=task_id, status=status))

#     def call_method(self, exchanger: 'Exchanger', task_id: str, method: str, params: dict[str, typing.Any]):
#         self.thread_pool.submit(ServiceServer._method_wrapper, task_id, exchanger, self.calls[method], **params)

#     TYPE_TO_YWPI = {
#         int: 'int',
#         str: 'str',
#         float: 'float'
#     }


class SimpleMethodExecuter:
    def __init__(self, registered_methods: dict[str, RegisteredMethod]) -> None:
        self.thread_pool = futures.ThreadPoolExecutor(max_workers=1)

        self.calls: dict[str, typing.Callable] = {}
        self.methods: list[hub_models.Method] = []
        self.method_input_dicts: dict[str, dict[str, InputTyping]] = {}

        # TaskID -> IOManager
        self._io_managers: dict[str, IOManager] = {}
        self._io_managers_lock = threading.Lock()

        for name, registered_method in registered_methods.items():
            self.methods.append(hub_models.Method(
                name=name,
                inputs=[
                    # hub_models.InputDescription(name=input_name, type=input.name)
                    # for input_name, input in registered_method.inputs.items()
                    hub_models.Field(
                        name=input_name,
                        type=hub_models.Type(
                            name=input.name,
                            args=[hub_models.Type(name=a.name) for a in input.args]
                        ),
                        description=input.description
                    )
                    for input_name, input in registered_method.inputs.items()
                ],
                outputs=[
                    hub_models.Field(
                        name=output_name,
                        type=hub_models.Type(
                            name=output.name,
                            args=[hub_models.Type(name=arg.name) for arg in output.args] if output.args is not None else []
                        )
                    )
                    for output_name, output in registered_method.outputs.items()
                ],
                description=registered_method.description,
                labels=list(map(lambda e: hub_models.Label(name=e), registered_method.labels)) if registered_method.labels else None,
                openai_schema=get_function_schema(registered_method.fn),
            ))
            self.calls[name] = registered_method.fn
            self.method_input_dicts[name] = registered_method.inputs

    def _perform_task_cleanup(self, task_id: str):
        logger.debug(f'Perform cleaup for task "{task_id}"')
        with self._io_managers_lock:
            self._io_managers.pop(task_id, None)

    @staticmethod
    def _method_wrapper(
        io_manager: IOManager,
        executer: 'SimpleMethodExecuter',
        method,
        kwargs
    ):
        final_outputs = None
        final_status = 'failed'
        error = None
        try:
            staticgenerator = isinstance(method, staticmethod) and inspect.isgeneratorfunction(method.__func__)
            if inspect.isgeneratorfunction(method) or staticgenerator:
                for outputs in method(**kwargs):
                    try:
                        io_manager.handle_outputs(outputs).result()
                    except TypeError as e:
                        logger.warning(f'Outputs serializations error: {e.args}')
            else:
                final_outputs = method(**kwargs)
            final_status = 'completed'
        except BaseException as e:
            # TODO: Move handling to distinct static method
            tb = traceback.format_exc()
            logger.warning(f"Method raise exception: {tb}")
            data = None
            try:
                serialization_model = get_exception_serialization_model(e)
                data = (
                    serialization_model
                    .model_validate(e, from_attributes=True)
                    .model_dump(mode='json')
                ) if serialization_model is not None else None
            except:
                logger.warning(f"Error while serializing error: {traceback.format_exc()}")

            error = hub_models.Error(
                type="MethodExecutionError",
                traceback=tb,
                data={
                    "source_exception_type": type(e).__name__,
                    "data": data
                }
            )
            final_status = "failed"
        finally:
            try:
                logger.debug(f'Start cleanup steps for task "{io_manager.task_id}"')
                io_manager.update_task_status(final_status, final_outputs, error=error).result()
                executer._perform_task_cleanup(io_manager.task_id)
            except:
                logger.error(f'Task cleanup steps error: {traceback.format_exc()}')

    def update_task_inputs(self, task_id: str, inputs: dict):
        with self._io_managers_lock:
            if task_id in self._io_managers:
                self._io_managers[task_id].update_inputs(inputs)
            else:
                logger.warning(f'Task "{task_id}" does not has io manager and could not be updated')

    def call_method(
        self,
        exchanger: 'Exchanger',
        task_id: str,
        method: str,
        inputs: dict[str, typing.Any],
        attachments: typing.MutableMapping[str, dict]
    ):
        # exchanger.call_update_task(hub_models.UpdateTaskRequest(id=task_id, status='started'))
        io_manager = IOManager(task_id, self.method_input_dicts[method], exchanger)

        # TODO: Move `handle_inputs` call to `_method_wrapper`
        # WHY: `handle_inputs` sometimes could perform long running job during convertation like file downloading
        params = io_manager.handle_inputs(inputs, attachments)

        return self.thread_pool.submit(
            SimpleMethodExecuter._method_wrapper,
            io_manager,
            self,
            self.calls[method],
            params
        )


# Communication level
class Exchanger:
    def __init__(self, input_channel, output_channel: Channel, service: 'SimpleMethodExecuter', agent_id = None) -> None:
        self.agent_id = agent_id
        self.incoming_requests: dict[str, futures.Future] = {}
        self.outgoings_requests: dict[str, futures.Future] = {}

        self.input_channel = input_channel
        self.output_channel = output_channel
        self.finish = futures.Future()
        self.outgoings_requests_lock = threading.Lock()
        self.service = service
        self.thr = threading.Thread(target=self._reader)
        self.thr.start()

    def _write_request_message(self, rpc: hub_pb2.Rpc, payload: str, reply_to: str):
        self.output_channel.push(
            hub_pb2.Message(
                reply_to=reply_to,
                request=hub_pb2.RequestMessage(
                    rpc=rpc,
                    payload=payload
                )
            )
        )

    def _write_response_message(
        self,
        reply_to: str,
        payload: str | None = None,
        error: str | None = 'undefined'
    ):
        args = { 'payload': payload } if payload else { 'error': error }
        logger.debug(f'[RPC] Write response MID: "{reply_to}" payload: "{payload}"')
        self.output_channel.push(
            hub_pb2.Message(
                reply_to=reply_to,
                response=hub_pb2.ResponseMessage(
                    **args
                )
            )
        )

    def _rpc_start_task(self, payload: hub_models.StartTaskRequest) -> hub_models.StartTaskResponse:
        status = 'started'
        self.service.call_method(self, payload.id, payload.method, payload.params, payload.attachments)
        return hub_models.StartTaskResponse(status=status)

    def _rpc_update_task(self, payload: hub_models.UpdateTaskRequest) -> hub_models.UpdateTaskResponse:
        """
        Hub can update task status (task abort) & inputs (streaming)
        """
        if payload.inputs is None:
            raise TypeError('inputs should be specified')

        if payload.status is not None:
            raise NotImplementedError()

        self.service.update_task_inputs(payload.id, payload.inputs)

    def _handle_request(self, reply_to: str, request: hub_pb2.RequestMessage, attachments: typing.MutableMapping[str, bytes]):
        try:
            if request.rpc == hub_pb2.Rpc.RPC_START_TASK:
                model_request = hub_models.StartTaskRequest.model_validate_json(request.payload)
                model_request.attachments = attachments
                response = self._rpc_start_task(model_request)
                self._write_response_message(
                    reply_to,
                    response.model_dump_json(),
                )
            elif request.rpc == hub_pb2.Rpc.RPC_ABORT_TASK:
                logger.warning(f'Method {hub_pb2.Rpc.Name(request.rpc)} not implemented')
            else:
                logger.warning(f'Method {hub_pb2.Rpc.Name(request.rpc)} not implemented')
        except BaseException as e:
            traceback.print_exc()
            self._write_response_message(
                reply_to,
                error=str(e)
            )

    def _reader(self):
        try:
            for message in self.input_channel:
                # Bug in aiochannel pkg: `def __aiter__(self) -> "Channel":` no type
                message: hub_pb2.Message
                attr = message.__getattribute__(message.WhichOneof('message'))

                if isinstance(attr, hub_pb2.ResponseMessage):
                    logger.debug(f'[RPC] Recieve response for MID: "{message.reply_to}" payload: "{attr.payload}"')
                    if message.reply_to in self.outgoings_requests:
                        with self.outgoings_requests_lock:
                            self.outgoings_requests.pop(message.reply_to).set_result(attr)
                    else:
                        logger.warning(f'[RPC] Recieved unexpected response MID: "{message.reply_to}"')
                elif isinstance(attr, hub_pb2.RequestMessage):
                    logger.debug(f'[RPC] Recieve rpc "{hub_pb2.Rpc.Name(attr.rpc)}" for "{message.reply_to}" payload: "{attr.payload}"')
                    self._handle_request(message.reply_to, attr, message.attachments)
                else:
                    logger.warning(f'Recieved unexpected message type {type(attr)}')
        except BaseException as e:
            self.finish.set_exception(e)
        else:
            self.finish.set_exception(RuntimeError("hub connection broken"))
        finally:
            with self.outgoings_requests_lock:
                [ f.set_exception(Exception()) for f in self.outgoings_requests.values() ]

    def call(self, rpc: hub_pb2.Rpc, payload: pydantic.BaseModel) -> futures.Future[hub_pb2.ResponseMessage]:
        reply_to = str(uuid.uuid4())
        logger.debug(f'[RPC] call (start): "{hub_pb2.Rpc.Name(rpc)}" MID: "{reply_to}" payload: "{payload}"')
        future = futures.Future()
        with self.outgoings_requests_lock:
            self.outgoings_requests[reply_to] = future
        self._write_request_message(rpc, payload, reply_to)
        logger.debug(f'[RPC] call: "{hub_pb2.Rpc.Name(rpc)}" MID: "{reply_to}" payload: "{payload}"')
        return future

    def call_register_agent(self, payload: hub_models.RegisterAgentRequest):
        return self.call(hub_pb2.Rpc.RPC_REGISTER_AGENT, payload.model_dump_json())

    def call_update_task(self, payload: hub_models.UpdateTaskRequest):
        return self.call(hub_pb2.Rpc.RPC_UPDATE_TASK, payload.model_dump_json())



grpc_channel_options = [
    ('grpc.max_send_message_length', settings.YWPI_GRPC_MAX_MESSAGE_SIZE),
    ('grpc.max_receive_message_length', settings.YWPI_GRPC_MAX_MESSAGE_SIZE),
    ('grpc.keepalive_time_ms', settings.YWPI_GRPC_KEEPALIVE_TIME_MS),
    ('grpc.keepalive_timeout_ms', settings.YWPI_GRPC_KEEPALIVE_TIMEOUT_MS),
]


def serve(
    id: str,
    name: str = 'Untitled',
    description: str = 'No description provided',
    project: str = settings.YWPI_PROJECT_NAME,
):
    service = SimpleMethodExecuter(REGISTERED_METHODS)
    with grpc.insecure_channel(settings.YWPI_HUB_HOST, options=grpc_channel_options) as grpc_channel:
        greeter_stub = hub_pb2_grpc.HubStub(grpc_channel)
        output_channel = Channel()
        response_iterator = greeter_stub.Connect(iter(output_channel))

        hello_message = hub_models.RegisterAgentRequest(
            id=id,
            name=name,
            project=project,
            description=description,
            methods=service.methods,
        )

        try:
            exchanger = Exchanger(response_iterator, output_channel, service)
            result = exchanger.call_register_agent(hello_message)
            good = result.result()
            if good.HasField('error'):
                logger.error(f'Agent register failed: {good.error}')
                raise Exception()

            logger.info(f'Connected to hub {settings.YWPI_HUB_HOST}')
            exchanger.finish.result()
        except KeyboardInterrupt:
            pass
        finally:
            output_channel.close()


def serve_with_reconnecting(*args, **kwargs):
    import time
    sleep_time = 1
    running_time = 0
    while True:
        try:
            start = time.time()
            serve(*args, **kwargs)
        except KeyboardInterrupt:
            return
        except BaseException as e:
            running_time = time.time() - start

        if running_time > 10:
            sleep_time = 1
        else:
            sleep_time = min(sleep_time * 2, 20)
        logger.warning(f'Connection broken. Try to reconnect after {sleep_time}')
        time.sleep(sleep_time)
        logger.info("Reconnecting ...")


def track(fn):
    def decorated(*args, **kwargs):
        with grpc.insecure_channel(settings.YWPI_HUB_HOST, options=grpc_channel_options) as grpc_channel:
            greeter_stub = hub_pb2_grpc.HubStub(grpc_channel)
            output_channel = Channel()
            response_iterator = greeter_stub.Connect(iter(output_channel))

            hello_message = hub_models.RegisterAgentRequest(
                id='id',
                name='name',
                project='project',
                description='description',
                methods=[],
            )

            try:
                exchanger = Exchanger(response_iterator, output_channel, None)
                result = exchanger.call_register_agent(hello_message)
                good = result.result()
                if good.HasField('error'):
                    logger.error(f'Agent register failed: {good.error}')
                    raise Exception()

                logger.info(f'Connected to hub {settings.YWPI_HUB_HOST}')

                task = exchanger.call(
                    hub_pb2.Rpc.RPC_START_TASK,
                    hub_models.StartTaskRequest(
                        id='',
                        method='test',
                        params={}
                    ).model_dump_json()
                )
                task = hub_models.StartTrackinTaskResponse.model_validate_json(task.result().payload)

                io_manager = IOManager(task.id, None, exchanger)
                SimpleMethodExecuter._method_wrapper(io_manager, None, fn, kwargs)

                # exchanger.finish.result()
            except KeyboardInterrupt:
                pass
            finally:
                output_channel.close()

    return decorated



from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class Rpc(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    RPC_UNDEFINED: _ClassVar[Rpc]
    RPC_REGISTER_AGENT: _ClassVar[Rpc]
    RPC_HEARTBEAT_AGENT: _ClassVar[Rpc]
    RPC_START_TASK: _ClassVar[Rpc]
    RPC_ABORT_TASK: _ClassVar[Rpc]
    RPC_UPDATE_TASK: _ClassVar[Rpc]
RPC_UNDEFINED: Rpc
RPC_REGISTER_AGENT: Rpc
RPC_HEARTBEAT_AGENT: Rpc
RPC_START_TASK: Rpc
RPC_ABORT_TASK: Rpc
RPC_UPDATE_TASK: Rpc

class RequestMessage(_message.Message):
    __slots__ = ("rpc", "payload")
    RPC_FIELD_NUMBER: _ClassVar[int]
    PAYLOAD_FIELD_NUMBER: _ClassVar[int]
    rpc: Rpc
    payload: str
    def __init__(self, rpc: _Optional[_Union[Rpc, str]] = ..., payload: _Optional[str] = ...) -> None: ...

class ResponseMessage(_message.Message):
    __slots__ = ("payload", "error")
    PAYLOAD_FIELD_NUMBER: _ClassVar[int]
    ERROR_FIELD_NUMBER: _ClassVar[int]
    payload: str
    error: str
    def __init__(self, payload: _Optional[str] = ..., error: _Optional[str] = ...) -> None: ...

class Message(_message.Message):
    __slots__ = ("reply_to", "request", "response", "attachments")
    class AttachmentsEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: bytes
        def __init__(self, key: _Optional[str] = ..., value: _Optional[bytes] = ...) -> None: ...
    REPLY_TO_FIELD_NUMBER: _ClassVar[int]
    REQUEST_FIELD_NUMBER: _ClassVar[int]
    RESPONSE_FIELD_NUMBER: _ClassVar[int]
    ATTACHMENTS_FIELD_NUMBER: _ClassVar[int]
    reply_to: str
    request: RequestMessage
    response: ResponseMessage
    attachments: _containers.ScalarMap[str, bytes]
    def __init__(self, reply_to: _Optional[str] = ..., request: _Optional[_Union[RequestMessage, _Mapping]] = ..., response: _Optional[_Union[ResponseMessage, _Mapping]] = ..., attachments: _Optional[_Mapping[str, bytes]] = ...) -> None: ...

class Error(_message.Message):
    __slots__ = ("type", "traceback", "data")
    TYPE_FIELD_NUMBER: _ClassVar[int]
    TRACEBACK_FIELD_NUMBER: _ClassVar[int]
    DATA_FIELD_NUMBER: _ClassVar[int]
    type: str
    traceback: str
    data: str
    def __init__(self, type: _Optional[str] = ..., traceback: _Optional[str] = ..., data: _Optional[str] = ...) -> None: ...

class PushTaskRequest(_message.Message):
    __slots__ = ("agent_id", "method", "params", "payload", "silent")
    AGENT_ID_FIELD_NUMBER: _ClassVar[int]
    METHOD_FIELD_NUMBER: _ClassVar[int]
    PARAMS_FIELD_NUMBER: _ClassVar[int]
    PAYLOAD_FIELD_NUMBER: _ClassVar[int]
    SILENT_FIELD_NUMBER: _ClassVar[int]
    agent_id: str
    method: str
    params: str
    payload: str
    silent: bool
    def __init__(self, agent_id: _Optional[str] = ..., method: _Optional[str] = ..., params: _Optional[str] = ..., payload: _Optional[str] = ..., silent: bool = ...) -> None: ...

class PushTaskResponse(_message.Message):
    __slots__ = ("task_id", "error")
    TASK_ID_FIELD_NUMBER: _ClassVar[int]
    ERROR_FIELD_NUMBER: _ClassVar[int]
    task_id: str
    error: Error
    def __init__(self, task_id: _Optional[str] = ..., error: _Optional[_Union[Error, _Mapping]] = ...) -> None: ...

class RunTaskResponse(_message.Message):
    __slots__ = ("task_id", "error", "outputs")
    TASK_ID_FIELD_NUMBER: _ClassVar[int]
    ERROR_FIELD_NUMBER: _ClassVar[int]
    OUTPUTS_FIELD_NUMBER: _ClassVar[int]
    task_id: str
    error: Error
    outputs: str
    def __init__(self, task_id: _Optional[str] = ..., error: _Optional[_Union[Error, _Mapping]] = ..., outputs: _Optional[str] = ...) -> None: ...

class GetAgentsListRequest(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class Input(_message.Message):
    __slots__ = ("name", "type")
    NAME_FIELD_NUMBER: _ClassVar[int]
    TYPE_FIELD_NUMBER: _ClassVar[int]
    name: str
    type: str
    def __init__(self, name: _Optional[str] = ..., type: _Optional[str] = ...) -> None: ...

class Method(_message.Message):
    __slots__ = ("name", "inputs")
    NAME_FIELD_NUMBER: _ClassVar[int]
    INPUTS_FIELD_NUMBER: _ClassVar[int]
    name: str
    inputs: _containers.RepeatedCompositeFieldContainer[Input]
    def __init__(self, name: _Optional[str] = ..., inputs: _Optional[_Iterable[_Union[Input, _Mapping]]] = ...) -> None: ...

class Agent(_message.Message):
    __slots__ = ("id", "name", "methods")
    ID_FIELD_NUMBER: _ClassVar[int]
    NAME_FIELD_NUMBER: _ClassVar[int]
    METHODS_FIELD_NUMBER: _ClassVar[int]
    id: str
    name: str
    methods: _containers.RepeatedCompositeFieldContainer[Method]
    def __init__(self, id: _Optional[str] = ..., name: _Optional[str] = ..., methods: _Optional[_Iterable[_Union[Method, _Mapping]]] = ...) -> None: ...

class GetAgentsListResponse(_message.Message):
    __slots__ = ("agents",)
    AGENTS_FIELD_NUMBER: _ClassVar[int]
    agents: _containers.RepeatedCompositeFieldContainer[Agent]
    def __init__(self, agents: _Optional[_Iterable[_Union[Agent, _Mapping]]] = ...) -> None: ...

class SubscribeOnAgentsRequest(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class SubscribeOnAgentsResponse(_message.Message):
    __slots__ = ("payload", "error")
    PAYLOAD_FIELD_NUMBER: _ClassVar[int]
    ERROR_FIELD_NUMBER: _ClassVar[int]
    payload: str
    error: Error
    def __init__(self, payload: _Optional[str] = ..., error: _Optional[_Union[Error, _Mapping]] = ...) -> None: ...



import dataclasses
import typing as t
import inspect
import pydantic
import pydantic_core

from ywpi import ytypes
# from ywpi.stream import Stream

TYPE_NAMES = {
    str: 'str',
    int: 'int',
    float: 'float',
    ytypes.Text: 'text',
    bytes: 'bytes',
    ytypes.Image: 'image',
    ytypes.Object: 'object',
    ytypes.Context: 'context',
    list: 'list',
    # Stream: 'stream'
}


SERIALIZERS: dict[t.Any, t.Callable] = {
    str: lambda v: str(v),
    int: lambda v: int(v),
}


def cvt_image(value, *args):
    ref = value['ref']
    tp = value['type']
    if tp != 'image':
        raise TypeError('ref image')
    return ytypes.Image()


def cvt_ref(value):
    ref = value['ref']
    return ytypes.Ref()

# def handle_stream(data: dict | list):
#     if isinstance(data, list):
#         return Stream(init_items=data)
#     return Stream(init_items=[data])

DESERIALIZERS: dict[t.Any, t.Callable] = {
    str: lambda v, _: v,
    int: lambda v, _: int(v),
    float: lambda v, _: float(v),
    ytypes.Image: cvt_image,
    ytypes.Text: lambda v, _: str(v),
    ytypes.Object: lambda v, _: ytypes.Object.model_validate(v),
    ytypes.Context: lambda v, _: ytypes.Context.model_validate(v),
    # Stream: handle_stream
}


# (From Type, To Type) -> Converter
TYPE_CONVERTERS = {
    (ytypes.Image, bytes): lambda v: bytes('image-data', encoding='utf-8'),
    (ytypes.Text, str): lambda v: str(v),
}


@dataclasses.dataclass
class Type:
    name: str
    tp: t.Any
    args: t.Optional[list['Type']] = dataclasses.field(default_factory=list)


@dataclasses.dataclass
class InputTyping:
    """
    Fields:
        name: type name
        target_tp: python object representation of type. (Actually type)
        json_repr: json representation of type. Usually it is dict with ywpi Ref 
    """
    name: str
    target_tp: t.Any
    source_tp: t.Any | None = None
    optional: bool = False
    json_repr: t.Optional[t.Union[int, str, float, dict, list]] = None
    args: t.Optional[list['Type']] = dataclasses.field(default_factory=list)
    description: t.Optional[str] = None


def handle_pydantic_field(field_info: pydantic.fields.FieldInfo):
    # Return tuple (description, optional)
    return (field_info.description, not isinstance(field_info.default, pydantic_core.PydanticUndefinedType))


def get_type_name(tp):
    if tp not in TYPE_NAMES:
        if issubclass(tp, pydantic.BaseModel):
            return tp.__name__
        raise TypeError('Type should be registered in `TYPE_NAMES`')

    return TYPE_NAMES[tp]


def get_input_dict(fn) -> dict[str, InputTyping]:
    """
    Retrieve typing from function arguments
    """
    inputs_dict: dict[str, InputTyping] = {  }

    for name, param in inspect.signature(fn).parameters.items():
        if param.annotation is inspect.Parameter.empty:
            raise TypeError(f'argument {name} must has annotation (before llm)')
        tp = param.annotation

        if isinstance(param.default, pydantic.fields.FieldInfo):
            description, optional = handle_pydantic_field(param.default)
        else:
            description, optional = None, param.default is not inspect.Parameter.empty

        if t.get_origin(tp) is t.Annotated:
            target_tp = tp.__origin__
            source_tp = tp.__metadata__[0]

            if source_tp not in DESERIALIZERS:
                raise KeyError(f'type {source_tp} has not got deserializer')

            if (source_tp, target_tp) not in TYPE_CONVERTERS:
                raise TypeError(f'no avalible conversation from {source_tp} to {target_tp}')

            if t.get_origin(source_tp) is not None:
                source_tp = t.get_origin(source_tp)

            inputs_dict[name] = InputTyping(
                name=get_type_name(source_tp),
                source_tp=source_tp,
                target_tp=target_tp,
                optional=optional,
                description=description
            )
        else:
            target_tp = tp
            # TODO: There probaly required deserialize subtypes (generic args)
            if t.get_origin(target_tp) is not None:
                target_tp = t.get_origin(target_tp)

            if target_tp in DESERIALIZERS:
                source_tp = target_tp
                inputs_dict[name] = InputTyping(
                    name=get_type_name(source_tp),
                    source_tp=target_tp,
                    target_tp=target_tp,
                    optional=optional,
                    description=description
                )
            elif issubclass(target_tp, pydantic.BaseModel):
                source_tp = target_tp

                args = pydantic._internal._generics.get_args(source_tp)

                type_args = []
                if len(args):
                    source_tp = pydantic._internal._generics.get_origin(source_tp)
                    arg0 = args[0]
                    type_args = [
                        Type(
                            name=get_type_name(arg0),
                            tp=arg0,
                        )
                    ]

                inputs_dict[name] = InputTyping(
                    name=get_type_name(source_tp),
                    source_tp=target_tp,
                    target_tp=target_tp,
                    args=type_args,
                    optional=optional,
                    description=description
                )
            else:
                input_type = handle_tp(tp)
                inputs_dict[name] = InputTyping(
                    name=get_type_name(input_type.tp),
                    source_tp=input_type.tp,
                    target_tp=input_type.tp,
                    optional=param.default is not inspect.Parameter.empty,
                    args=input_type.args
                )
                # raise KeyError(f'type {tp} has not got deserializer')
    return inputs_dict


def handle_arg(data: dict, tp: Type, attachments: t.MutableMapping[str, bytes]):
    return DESERIALIZERS[tp.tp](data, attachments)


def handle_args(data: dict, inputs: dict[str, InputTyping], attachments: t.MutableMapping[str, dict], ctx: dict = {}):
    """
    Description:
        Convert `data` in Referenced JSON format to Python dictionary using `schema`.
        This process include:
            - Type conversation during `typing.Annotated` annotation
            - File downloading (If ywpi reference used)
            - Stream creation
        Create streams if required.

        Binary content parsing: { "$attachment": "<attachment key>" }

    Returns:
        Converted data.
    """
    result_args = {}
    for name, input in inputs.items():
        if name not in data:
            # TODO: Ignore if input type is `ywpi.Stream`
            if input.optional:
                continue
            raise KeyError(f'argument {name} does not present in inputs')
        raw_value = data[name]

        source_tp = t.get_origin(input.source_tp) if t.get_origin(input.source_tp) is not None else input.source_tp

        if source_tp in DESERIALIZERS:
            value = DESERIALIZERS[source_tp](raw_value, attachments)
        elif issubclass(input.source_tp, pydantic.BaseModel):
            context = { "attachments": attachments }
            value = input.source_tp.model_validate(raw_value, context=context)
        elif input.source_tp is list:
            if type(raw_value) is not list:
                raise TypeError('input type is not list')
            value = [handle_arg(rv, input.args[0], attachments) for rv in raw_value]
        elif input.source_tp is bytes:
            value = attachments[data[name]["$attachment"]]
        else:
            raise RuntimeError(f'no handle for type {input.source_tp}')

        if input.source_tp is not input.target_tp:
            value = TYPE_CONVERTERS[(input.source_tp, input.target_tp)](value)

        result_args[name] = value
    return result_args


# import ywpi

# # def fn(text: str, image: t.Annotated[bytes, ywpi.Image]): pass
# def fn(text: str, image: t.Annotated[bytes, ywpi.Image] = None): pass

# # def fn(text: str, image: t.Annotated[bytes, ywpi.Image], thr: int = 1): pass


# inputs_dict = get_input_dict(fn)
# print(inputs_dict)

# res = handle_args(
#     {
#         'text': 'string',
#         'image': {
#             'ref': 46527,
#             'type': 'image',
#             'href': 'https://drive.ywpi.ru/o/46527'
#         }
#     },
#     # {
#     #     'text': InputTyping(name='str', target_tp=str, source_tp=str),
#     #     'image': InputTyping(name='image', target_tp=bytes, source_tp=ywpi.Image)
#     # }
#     inputs_dict
# )
# print(res)

# # print(type_hints['text'].__origin__, type_hints['text'].__metadata__, type_hints['text'])
# # print(t.get_origin(t.Annotated[str, ywpi.Text]) is t.Annotated)


def handle_tp(tp: t.Any) -> Type:
    args = t.get_args(tp)
    orig = t.get_origin(tp) if args else tp

    if orig in (list, set):
        out = Type(name=orig.__name__, tp=orig)
        if args:
            out.args = [
                handle_tp(args[0])
            ]
        return out
    else:
        return Type(name=orig.__name__, tp=orig)


def handle_ret(fn):
    tp = inspect.signature(fn).return_annotation

    if not t.get_args(tp) and issubclass(inspect.Parameter.empty, tp): return

    if inspect.isclass(tp) and issubclass(tp, pydantic.BaseModel):
        return {
            k: handle_tp(v.annotation)
            for k, v in tp.model_fields.items()
        }

    return handle_tp(tp)


def get_output_dict(fn):
    t = handle_ret(fn)
    if t is not None:
        if isinstance(t, dict):
            return t

        if t.tp in (list, set) and t.args is not None and len(t.args) > 0:
            return {
                '__others__': t.args[0]
            }
        return {}
    else:
        return {}




import typing as t

T = t.TypeVar('T')


class Stream(t.Generic[T]):
    def __init__(
        self,
        get_item_callback: t.Callable[[], T],
        init_items: list | None = None
    ):
        self._init_items = init_items
        self._get_item_callback = get_item_callback

    def __next__(self) -> T:
        return self._get_item_callback()

    def __iter__(self):
        return self


__all__ = (
    'Stream',
)



import ywpi

def type_name_to_python_type(name: str):
    if name in ('context', 'object'):
        return f'ywpi.{name.capitalize()}'

    return name


base_model_template = """
class {class_name}(pydantic.BaseModel):
{class_content}
"""


base_agent_template = """
class {class_name}(pydantic.BaseModel):
{class_content}
"""


def get_additional_definition(type_name):
    if type_name == 'CustomMessage':
        return base_model_template.format(
            class_name='CustomMessage',
            class_content='\n'.join([
                '    id: int',
                '    images: list[str]'
            ])
        )


def generate_stub_file_content(methods: list[ywpi.Method]):
    methods = ywpi.get_methods()
    imports = ["import ywpi", "import pydantic"]
    type_definitions = {}
    definitions = []
    for m in methods:
        inputs = []
        for i in m.inputs:
            inputs.append(f"{i.name}: {type_name_to_python_type(i.type)}")
            # if i.type not in type_definitions:
            #     definition = get_additional_definition(i.type)
            #     if definition is not None:
            #         type_definitions[i.type] = definition
        definitions.append(f"def {m.name}({ ', '.join(inputs) }): ...")

    return "\n\n".join(imports + list(type_definitions.values()) + definitions) + '\n'



import typing as t

import pydantic


def get_exception_serialization_model(e: BaseException) -> type[pydantic.BaseModel] | None:
    exception_tp = type(e)
    type_hints = t.get_type_hints(exception_tp)
    model_tp = type_hints.get("serialization_model", None)
    if model_tp is not None and issubclass(model_tp, pydantic.BaseModel):
        return model_tp




import pydantic
import typing

from abc import ABC



sample_ref = {
    "__adapter__": "YwpiDriveAdapter",
    "type": "",
    "ref": "",
}

class DriveRef(pydantic.BaseModel):
    adapter: str = pydantic.Field(alias='__adapter__')
    type: str
    ref: str

class AbstractDriveAdapter:
    def materialize_ref(self, ref: DriveRef) -> typing.Any: pass
    def referencify_object(self, obj: typing.Any) -> DriveRef: pass


def handle_args(data: dict, drive_adapter: AbstractDriveAdapter):
    for name, value in data:
        if isinstance(value, dict):
            drive_adapter.materialize_ref(value)
        else:
            pass



# Multiple drive targets & sources
# Default ywpi drive




from ywpi.hub_pb2_grpc import HubStub




class Method:
    def call(self, *args, **params):
        if len(args):
            raise RuntimeError('call method can not recieve positional arguments')

    def to_langchain_tool(self):
        return None


class Client:
    def __init__(self, host):
        pass

    def get_method(self) -> Method:
        pass

    def get_methods(
        self,
        tag: list[str] | str = None,
        agent_class: list[str] | str = None,
        agent_id: list[str] | str = None,
    ) -> list[Method]:
        pass

    def call_method(self):
        pass


client = Client()
tools = [ m.to_langchain_tool() for m in client.get_methods(agent_class='MyWorker') ]




from .decorators import *
from .main import *
from .ytypes import *
from .serialization import serial, deserial
from .stream import *
from .api import *

# Additonal document platform specific functionality
from .additional_types import *
from .stub.getattr import stub_getattr
from .hub_models import YwpiException, MethodCallError, MethodExecutionError



import pydantic
import typing as t
import inspect
from collections import defaultdict
import re

# This code is a copy-paste from ollama pgk
# There should be one more implementation from autogen

def _parse_docstring(doc_string: t.Union[str, None]) -> dict[str, str]:
    parsed_docstring = defaultdict(str)
    if not doc_string:
        return parsed_docstring

    key = str(hash(doc_string))
    for line in doc_string.splitlines():
        lowered_line = line.lower().strip()
        if lowered_line.startswith('args:'):
            key = 'args'
        elif lowered_line.startswith(('returns:', 'yields:', 'raises:')):
            key = '_'
        else:
            # maybe change to a list and join later
            parsed_docstring[key] += f'{line.strip()}\n'

    last_key = None
    for line in parsed_docstring['args'].splitlines():
        line = line.strip()
        if ':' in line:
            # Split the line on either:
            # 1. A parenthetical expression like (integer) - captured in group 1
            # 2. A colon :
            # Followed by optional whitespace. Only split on first occurrence.
            parts = re.split(r'(?:\(([^)]*)\)|:)\s*', line, maxsplit=1)

            arg_name = parts[0].strip()
            last_key = arg_name

            # Get the description - will be in parts[1] if parenthetical or parts[-1] if after colon
            arg_description = parts[-1].strip()
            if len(parts) > 2 and parts[1]:  # Has parenthetical content
                arg_description = parts[-1].split(':', 1)[-1].strip()

            parsed_docstring[last_key] = arg_description

        elif last_key and line:
            parsed_docstring[last_key] += ' ' + line

    return parsed_docstring


def get_function_schema(func: t.Callable) -> dict:
    doc_string_hash = str(hash(inspect.getdoc(func)))
    parsed_docstring = _parse_docstring(inspect.getdoc(func))

    schema = type(
        func.__name__,
        (pydantic.BaseModel,),
        {
            '__annotations__': {
                k: v.annotation if v.annotation != inspect._empty else str for k, v in inspect.signature(func).parameters.items()
            },
            '__signature__': inspect.signature(func),
            '__doc__': parsed_docstring[doc_string_hash],
        },
    ).model_json_schema()

    for k, v in schema.get('properties', {}).items():
        # If type is missing, the default is string
        types = { t.get('type', 'string') for t in v.get('anyOf')} if 'anyOf' in v else {v.get('type', 'string') }
        if 'null' in types:
            schema['required'].remove(k)
            types.discard('null')

        schema['properties'][k] = {
            'description': parsed_docstring[k],
            'type': ', '.join(types),
        }

    return {
        "type": "function",
        "function": {
            "name": func.__name__,
            "description": schema.pop("description", ""),
            "parameters": {
                **schema
            }
        }
    }



# Copied from packages/ywpi_hub/src/ywpi_hub/hub_models.py DO NOT EDIT!

import typing as t

import pydantic


class InputDescription(pydantic.BaseModel):
    name: str
    type: str


class Type(pydantic.BaseModel):
    name: str
    args: t.Optional[list['Type']] = None


class Field(pydantic.BaseModel):
    name: str
    type: Type
    description: t.Optional[str] = None


class Label(pydantic.BaseModel):
    name: str
    value: t.Optional[str] = None


class Subscribtion(pydantic.BaseModel):
    selector: dict


class Method(pydantic.BaseModel):
    name: str
    inputs: list[Field]
    outputs: list[Field]
    description: t.Optional[str] = None
    labels: t.Optional[list[Label]] = None
    subscribtions: t.Optional[list[Subscribtion]] = None
    openai_schema: t.Optional[dict] = None # OpenAI compatible function schema


class RegisterAgentRequest(pydantic.BaseModel):
    id: str
    name: str
    project: t.Optional[str] = None
    description: t.Optional[str] = None
    methods: list[Method]


class RegisterAgentResponse(pydantic.BaseModel):
    pass


class StartTaskRequest(pydantic.BaseModel):
    id: str
    method: str
    params: dict[str, t.Any] = { }
    attachments: t.MutableMapping[str, bytes] = { } # Type from protobuf


class StartTaskResponse(pydantic.BaseModel):
    status: str = 'success'


class StartTrackinTaskResponse(pydantic.BaseModel):
    id: str

ErrorTypeT = t.Union[t.Literal['MethodExecutionError', 'MethodCallError'], str]

class Error(pydantic.BaseModel):
    type: t.Union[t.Literal['MethodExecutionError', 'MethodCallError'], str] # Exception type
    traceback: t.Optional[str] = None
    data: t.Optional[dict] = None


class UpdateTaskRequest(pydantic.BaseModel):
    """
    `error` field ignored if `status != failed`
    """
    id: str
    status: t.Optional[str] = None
    outputs: t.Optional[dict] = None
    error: t.Optional[Error] = None


class UpdateTaskResponse(pydantic.BaseModel):
    pass


# Exceptions (should be in distinct file)


class YwpiException(BaseException):
    type: str # TODO: Remove `type` from exception and create `UnknownException` class with fiedl type
    traceback: t.Optional[str] = None
    data: t.Optional[dict] = None

    _subclasses: dict[str, t.Type['YwpiException']] = {}

    """
    Base serializable exception that contains json serializable `data`.
    """
    def __init__(
        self,
        *args,
        type: str = "YwpiException", # RuntimeError, TypeError
        data: t.Optional[dict] = None,
        traceback: str = None
    ):
        super().__init__(*args)
        self.type = type
        self.data = data
        self.traceback = traceback

    @classmethod
    def from_error(cls, error: Error):
        if error.type in YwpiException._subclasses:
            return YwpiException._subclasses[type](type=error.type, data=error.data, traceback=error.traceback)

        return YwpiException(type=error.type, data=error.data, traceback=error.traceback)

    def to_error(self) -> Error:
        return Error(
            type=self.type,
            data=self.data,
            traceback=self.traceback
        )

    @classmethod
    def from_data(cls, type: ErrorTypeT, data: t.Optional[dict] = None, traceback: t.Optional[str] = None):
        if type in YwpiException._subclasses:
            return YwpiException._subclasses[type](type=type, data=data, traceback=traceback)

        return YwpiException(type=type, data=data, traceback=traceback)

    def __init_subclass__(cls, **kwargs):
        # Collect all subclass exceptions
        YwpiException._subclasses[cls.__name__] = cls


class AgentNotFoundError(YwpiException): ...


class MethodNotFoundError(YwpiException): ...


class MethodCallError(YwpiException):
    """Base class for communication errors"""


class MethodExecutionError(YwpiException):
    """
    Base class for user defined method exceptions. It is container for exceptions that method raised.

    Why is this class container for exception?
    In both cases information about error inheritance will be losted, but in encapsulation
    case (current implementation) we got information about **place** where it was raised - in user method.
    In simple inheritance from `MethodExecutionError` we should save additional property about inheritance fact.
    """

    source_exception_type: str
    data: t.Optional[dict] = None # User defined optional payload

    def __init__(self, *args, type = "MethodExecutionError", data = None, traceback = None):
        super().__init__(*args, type=type, data=data, traceback=traceback)
        self.source_exception_type = data["source_exception_type"]
        self.data = data["data"]



import os


YWPI_HUB_HOST = os.getenv('YWPI_HUB_HOST', 'localhost:50051')

# By default name of the working directory
env_project_name = os.getenv('YWPI_PROJECT_NAME')
YWPI_PROJECT_NAME = env_project_name if env_project_name is not None else os.path.split(os.getcwd())[1]


# gRPC settings
YWPI_GRPC_MAX_MESSAGE_SIZE = int(os.getenv('YWPI_GRPC_MAX_MESSAGE_SIZE', 1024 * 1024 * 1024)) # Default 1G

# Keepalive required for long calling `Connect` method (client freeze while server disconnect agent)
# As mentioned in docs https://grpc.io/docs/guides/keepalive/ , client has infinite timeout by default
YWPI_GRPC_KEEPALIVE_TIME_MS = int(os.getenv('YWPI_GRPC_KEEPALIVE_TIME_MS', 1_200_000)) # Default 20 min
YWPI_GRPC_KEEPALIVE_TIMEOUT_MS = int(os.getenv('YWPI_GRPC_KEEPALIVE_TIMEOUT_MS', 20_000)) # Default 20 sec



import grpc

from . import base
from ywpi import settings
from ywpi import hub_pb2
from ywpi import hub_pb2_grpc

grpc_channel: grpc.aio.Channel | None = None
hub_stub: hub_pb2_grpc.HubStub | None = None


async def _init_stub():
    global hub_stub
    global grpc_channel
    if hub_stub is None:
        grpc_channel = grpc.aio.insecure_channel(settings.YWPI_HUB_HOST)
        hub_stub = hub_pb2_grpc.HubStub(grpc_channel)


class Method(base.Method):
    async def __call__(self, *args, **kwds):
        await _init_stub()
        return self._handle_response(
            await hub_stub.RunTask(
                self._create_request(*args, **kwds)
            )
        )


async def get_methods() -> list[Method]:
    await _init_stub()
    result = []
    response: hub_pb2.GetAgentsListResponse = await hub_stub.GetAgentsList(hub_pb2.GetAgentsListRequest())

    for agent in response.agents:
        for method in agent.methods:
            result.append(Method(
                agent.id,
                method.name,
                method.inputs
            ))
    
    return result


async def get_method(agent: str, name: str) -> Method:
    await _init_stub()
    response: hub_pb2.GetAgentsListResponse = await hub_stub.GetAgentsList(hub_pb2.GetAgentsListRequest())

    for a in response.agents:
        if a.id != agent:
            continue

        for m in a.methods:
            if m.name != name:
                continue
            return Method(
                a.id,
                m.name,
                m.inputs
            )
    raise RuntimeError('Method not found')



#################################################################
import asyncio
from aiochannel import Channel
from ywpi.logger import logger
import typing as t


class Connection:
    def __init__(self,
        input_channel: t.AsyncIterable[hub_pb2.Message],
        output_channel: Channel
    ) -> None:
        self.incoming_requests: dict[str, asyncio.Future] = {}
        self.outgoings_requests: dict[str, asyncio.Future] = {}

        self.input_channel = input_channel
        self.output_channel = output_channel
        self.finish = asyncio.Future
        self.outgoings_requests_lock = asyncio.Lock()

    async def start(self):
        self._reader_task = asyncio.create_task(self._reader())

    async def _write_request_message(self, rpc: hub_pb2.Rpc, payload: str, reply_to: str):
        self.output_channel.put_nowait(
            hub_pb2.Message(
                reply_to=reply_to,
                request=hub_pb2.RequestMessage(
                    rpc=rpc,
                    payload=payload
                )
            )
        )

    async def _write_response_message(
        self,
        reply_to: str,
        payload: str | None = None,
        error: str | None = 'undefined'
    ):
        args = { 'payload': payload } if payload else { 'error': error }
        logger.debug(f'[RPC] Write response MID: "{reply_to}" payload: "{payload}"')
        self.output_channel.put_nowait(
            hub_pb2.Message(
                reply_to=reply_to,
                response=hub_pb2.ResponseMessage(
                    **args
                )
            )
        )

    async def _reader(self):
        try:
            async for message in self.input_channel:
                attr = message.__getattribute__(message.WhichOneof('message'))

                if isinstance(attr, hub_pb2.ResponseMessage):
                    logger.debug(f'[RPC] Recieve response for MID: "{message.reply_to}" payload: "{attr.payload}"')
                    if message.reply_to in self.outgoings_requests:
                        self.outgoings_requests.pop(message.reply_to).set_result(attr)
                    else:
                        logger.warning(f'[RPC] Recieved unexpected response MID: "{message.reply_to}"')
                elif isinstance(attr, hub_pb2.RequestMessage):
                    logger.debug(f'[RPC] Recieve rpc "{hub_pb2.Rpc.Name(attr.rpc)}" for "{message.reply_to}" payload: "{attr.payload}"')
                    self._handle_request(message.reply_to, attr)
                else:
                    logger.warning(f'Recieved unexpected message type {type(attr)}')
        except BaseException as e:
            self.finish.set_exception(e)
        else:
            self.finish.set_exception(Exception())
        finally:
            for f in self.outgoings_requests.values():
                f.set_exception(Exception())


    async def _handle_request(self, reply_to: str, request: hub_pb2.RequestMessage):
        logger.warning('"_handle_request" method not implemented')

from ywpi import hub_models
import pydantic
import uuid
import traceback
import inspect


class MethodExecuter:
    def __init__(self):
        self._executions: Channel = Channel()

    async def _worker_loop(self):
        async for execution in self._executions:
            pass

    @staticmethod
    def _method_wrapper(
        execution_id: str,
        method: t.Callable,
        params: dict[str, t.Any],
        update_task: t.Callable[[hub_models.UpdateTaskRequest], t.Awaitable]
    ):
        final_outputs = None
        final_status = 'failed'
        try:
            staticgenerator = isinstance(method, staticmethod) and inspect.isgeneratorfunction(method.__func__)
            if inspect.isgeneratorfunction(method) or staticgenerator:
                for outputs in method(**kwargs):
                    try:
                        io_manager.handle_outputs(outputs).result()
                    except TypeError as e:
                        logger.warning(f'Outputs serializations error: {e.args}')
            else:
                final_outputs = method(**kwargs)
            final_status = 'completed'
        except BaseException as e:
            logger.warning(f'Method raise exception: {traceback.format_exc()}')
            final_status = 'failed'
        finally:
            try:
                logger.debug(f'Start cleanup steps for task "{io_manager.task_id}"')
                io_manager.update_task_status(final_status, final_outputs).result()
                executer._perform_task_cleanup(io_manager.task_id)
            except:
                logger.error(f'Task cleanup steps error: {traceback.format_exc()}')

    def execute_method(self,
        execution_id: str,
        method_name: str,
        params: dict[str, t.Any],
        update_task: t.Callable[[hub_models.UpdateTaskRequest], t.Awaitable]
    ):
        self._executions.put_nowait()


class Service(Connection):
    def __init__(self,
        input_channel: t.AsyncIterable[hub_pb2.Message],
        output_channel: Channel,
        method_executer: MethodExecuter
    ):
        super().__init__(input_channel, output_channel)
        self._method_executer = method_executer

    async def _handle_request(self, reply_to: str, request: hub_pb2.RequestMessage):
        try:
            if request.rpc == hub_pb2.Rpc.RPC_START_TASK:
                response = await self._rpc_start_task(
                    hub_models.StartTaskRequest.model_validate_json(request.payload)
                )
                await self._write_response_message(
                    reply_to,
                    response.model_dump_json(),
                )
            elif request.rpc == hub_pb2.Rpc.RPC_ABORT_TASK:
                logger.warning(f'Method {hub_pb2.Rpc.Name(request.rpc)} not implemented')
            else:
                logger.warning(f'Method {hub_pb2.Rpc.Name(request.rpc)} not implemented')
        except BaseException as e:
            traceback.print_exc()
            self._write_response_message(
                reply_to,
                error=str(e)
            )

    async def _rpc_start_task(self, payload: hub_models.StartTaskRequest) -> hub_models.StartTaskResponse:
        status = 'started'
        self._method_executer.execute_method(self.call_update_task, payload.id, payload.method, payload.params)
        return hub_models.StartTaskResponse(status=status)

    async def _rpc_update_task(self, payload: hub_models.UpdateTaskRequest) -> hub_models.UpdateTaskResponse:
        """
        Hub can update task status (task abort) & inputs (streaming)
        """
        raise NotImplementedError()

    async def call(self, rpc: hub_pb2.Rpc, payload: pydantic.BaseModel) -> asyncio.Future[hub_pb2.ResponseMessage]:
        reply_to = str(uuid.uuid4())
        # logger.debug(f'[RPC] call (start): "{hub_pb2.Rpc.Name(rpc)}" MID: "{reply_to}" payload: "{payload}"')
        future = asyncio.Future()
        with self.outgoings_requests_lock: # Lock does bot required
            self.outgoings_requests[reply_to] = future
        await self._write_request_message(rpc, payload, reply_to)
        logger.debug(f'[RPC] call: "{hub_pb2.Rpc.Name(rpc)}" MID: "{reply_to}" payload: "{payload}"')
        return future

    async def call_register_agent(self, payload: hub_models.RegisterAgentRequest):
        return await self.call(hub_pb2.Rpc.RPC_REGISTER_AGENT, payload.model_dump_json())

    async def call_update_task(self, payload: hub_models.UpdateTaskRequest):
        return await self.call(hub_pb2.Rpc.RPC_UPDATE_TASK, payload.model_dump_json())
